﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Encodings.BinaryVectorEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Problems.NK {

  [Item("SortedRandomInteractionsInitializer", "Randomly assigned interactions across all bits and sorts components according to bit affinity.")]
  [StorableClass]
  public class SortedRandomInteractionsInitializer : ParameterizedNamedItem, IInteractionInitializer {

    public ConstrainedValueParameter<IBinaryVectorComparer> ComparerParameter {
      get { return (ConstrainedValueParameter<IBinaryVectorComparer>)Parameters["Comparer"]; }
    }

    public IBinaryVectorComparer Comparer {
      get { return ComparerParameter.Value; }
    }
    
    [StorableConstructor]
    protected SortedRandomInteractionsInitializer(bool serializing) : base(serializing) { }
    protected SortedRandomInteractionsInitializer(SortedRandomInteractionsInitializer original, Cloner cloner) : base(original, cloner) { }
    public SortedRandomInteractionsInitializer() {
      AddComparerParameter();
    }

    private void AddComparerParameter() {
      Parameters.Add(new ConstrainedValueParameter<IBinaryVectorComparer>("Comparer", "Comparison for sorting of component functions"));
      foreach (var comparer in ApplicationManager.Manager.GetInstances<IBinaryVectorComparer>())
        ComparerParameter.ValidValues.Add(comparer);
      ComparerParameter.Value = ComparerParameter.ValidValues.First(v => v is AverageBitBinaryVectorComparer);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SortedRandomInteractionsInitializer(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      try {
        IBinaryVectorComparer comparer = Comparer;
      } catch {
        Parameters.Remove("Comparer");
        AddComparerParameter();
      }
    }

    #region IInteractionInitializer Members
    public BoolMatrix InitializeInterations(int length, int nComponents, int nInteractions, IRandom random) {
      BinaryVector[] components = Enumerable.Range(0, nComponents).Select(i => new BinaryVector(length)).ToArray();
      for (int c = 0; c < components.Length; c++) {
        var indices = Enumerable.Range(0, length).ToList();
        if (indices.Count > c) {
          indices.RemoveAt(c);
          components[c][c] = true;          
        }
        while (indices.Count > nInteractions) {
          indices.RemoveAt(random.Next(indices.Count));
        }
        foreach (var i in indices) {
          components[c][i] = true;          
        }
      }
      BoolMatrix m = new BoolMatrix(length, nComponents);
      foreach (var c in components.OrderBy(v => v, Comparer).Select((v, j) => new {v, j})) {
        for (int i = 0; i<c.v.Length; i++) {
          m[i, c.j] = c.v[i];
        }    
      }
      return m;
    }
    #endregion
  }
}
